-- Item indicator for Resident Evil 4
-- Author: chenstack
-- Version: 1.2.2
-- Functions: Mark item, box, treasure, blue medallions and Clockwork Castellan on screen, and experimental auto pick up.

local characterManager = sdk.get_managed_singleton("chainsaw.CharacterManager")
local ItemIDDefine_typedef = sdk.find_type_definition("chainsaw.ItemIDDefine")
local isKeyItem = ItemIDDefine_typedef:get_method("isKeyItem")
local isTreasureItem = ItemIDDefine_typedef:get_method("isTreasureItem")
local isInventoryItem = ItemIDDefine_typedef:get_method("isInventoryItem")
local DropItem_typedef = sdk.find_type_definition("chainsaw.DropItem")
local GmSmoothWoodBox_typedef = sdk.find_type_definition("chainsaw.GmSmoothWoodBox")
local GmShootObjBase_typedef = sdk.find_type_definition("chainsaw.GmShootObjBase")
local GmBlueCoin_typedef = sdk.find_type_definition("chainsaw.GmBlueCoin")
local GmSrRamon_typedef = sdk.find_type_definition("chainsaw.GmSrRamon")

local itemManager = sdk.get_managed_singleton("chainsaw.ItemManager")
local itemMessageManager = itemManager:get_ItemMessageManager()
-- local sceneManager = sdk.get_native_singleton("via.SceneManager")
-- local SceneManager_typedef = sdk.find_type_definition("via.SceneManager")

local key_bind = (function ()
    local keyboard_singleton = sdk.get_native_singleton("via.hid.Keyboard")
    local keyboard_typedef = sdk.find_type_definition("via.hid.Keyboard")
    local keyboardkey_typedef = sdk.find_type_definition("via.hid.KeyboardKey")
    local gamePad_singleton = sdk.get_native_singleton("via.hid.GamePad")
    local gamePad_typedef = sdk.find_type_definition("via.hid.GamePad")
    local mouse_singleton = sdk.get_native_singleton("via.hid.Mouse")
    local mouse_typedef = sdk.find_type_definition("via.hid.Mouse")

    local TRIGGER_TYPE_KEYBOARD = 1
    local TRIGGER_TYPE_CONTROLLER = 2
    local TRIGGER_TYPE_MOUSE = 3

    local cfg = {
        long_press_seconds = 0.2,
        -- next time can trigger delta seconds
        long_press_suppress = 0.5,
    }

    local function get_key_code(keyname)
        local field = keyboardkey_typedef:get_field(keyname)
        return field and field:get_data(nil)
    end

    -- return keyname, keycode
    local function try_get_keycode(keyname, default_name)
        keyname = string.gsub(keyname, "%s+", "")
        keyname = string.gsub(keyname, "^%l", string.upper)
        local keycode = get_key_code(keyname)
        if keycode then
            return keyname, keycode
        end
        return default_name, get_key_code(default_name)
    end

    -- inputs
    local trigger_type_choices = { "Keyboard", "Controller", "Mouse" }
    local trigger_type_choices_zh = { "键盘", "手柄", "鼠标" }

    local keyboard_time = {}
    local keyboard_long_press_next = 0
    local function is_keyboard_long_press(kb, code)
        local now = os.clock()
        if kb:isDown(code) then
            if keyboard_time[code] == nil then
                keyboard_time[code] = now
            elseif now > keyboard_long_press_next and now - keyboard_time[code] >= cfg.long_press_seconds then
                keyboard_time[code] = nil
                keyboard_long_press_next = now + cfg.long_press_suppress
                return true
            end
        else
            keyboard_time[code] = nil
        end
        return false
    end

    local pad_time = {}
    local pad_long_press_next = 0
    local pad_isDown = sdk.find_type_definition("via.hid.GamePadDevice"):get_method("isDown(via.hid.GamePadButton)")
    local function is_pad_long_press(pad, code)
        local now = os.clock()
        if pad_isDown:call(pad, code) then
            if pad_time[code] == nil then
                pad_time[code] = now
            elseif now > pad_long_press_next and now - pad_time[code] >= cfg.long_press_seconds then
                pad_time[code] = nil
                pad_long_press_next = now + cfg.long_press_suppress
                return true
            end
        else
            pad_time[code] = nil
        end
        return false
    end

    local mouse_time = {}
    local mouse_long_press_next = 0
    local mouse_choices = { "Left", "Right", "Center", "EX0", "EX1" }
    local mouse_choices_zh = { "左", "右", "中", "后退", "前进" }
    local mouse_values = { 1, 2, 4, 32, 64 }
    local function is_mouse_long_press(mouse, code)
        local now = os.clock()
        if mouse:isDown(code) then
            if mouse_time[code] == nil then
                mouse_time[code] = now
            elseif now > mouse_long_press_next and now - mouse_time[code] >= cfg.long_press_seconds then
                mouse_time[code] = nil
                mouse_long_press_next = now + cfg.long_press_suppress
                return true
            end
        else
            mouse_time[code] = nil
        end
        return false
    end

    local key_code_to_name = {
        [0x00] = "None",
        [0x01] = "LButton",
        [0x02] = "RButton",
        [0x03] = "Cancel",
        [0x04] = "MButton",
        [0x05] = "XButton1",
        [0x06] = "XButton2",
        [0x08] = "Back",
        [0x09] = "Tab",
        [0x0C] = "Clear",
        [0x0D] = "Enter",
        [0x10] = "Shift",
        [0x11] = "Control",
        [0x12] = "Menu",
        [0x13] = "Pause",
        [0x14] = "Capital",
        [0x15] = "Kana",
        [0x17] = "Junja",
        [0x18] = "Final",
        [0x19] = "Hanja",
        [0x1B] = "Escape",
        [0x1C] = "Convert",
        [0x1D] = "NonConvert",
        [0x1E] = "Accept",
        [0x1F] = "ModeChange",
        [0x20] = "Space",
        [0x21] = "Prior",
        [0x22] = "Next",
        [0x23] = "End",
        [0x24] = "Home",
        [0x25] = "Left",
        [0x26] = "Up",
        [0x27] = "Right",
        [0x28] = "Down",
        [0x29] = "Select",
        [0x2A] = "Print",
        [0x2B] = "Execute",
        [0x2C] = "SnapShot",
        [0x2D] = "Insert",
        [0x2E] = "Delete",
        [0x2F] = "Help",
        [0x30] = "Alpha0",
        [0x31] = "Alpha1",
        [0x32] = "Alpha2",
        [0x33] = "Alpha3",
        [0x34] = "Alpha4",
        [0x35] = "Alpha5",
        [0x36] = "Alpha6",
        [0x37] = "Alpha7",
        [0x38] = "Alpha8",
        [0x39] = "Alpha9",
        [0x41] = "A",
        [0x42] = "B",
        [0x43] = "C",
        [0x44] = "D",
        [0x45] = "E",
        [0x46] = "F",
        [0x47] = "G",
        [0x48] = "H",
        [0x49] = "I",
        [0x4A] = "J",
        [0x4B] = "K",
        [0x4C] = "L",
        [0x4D] = "M",
        [0x4E] = "N",
        [0x4F] = "O",
        [0x50] = "P",
        [0x51] = "Q",
        [0x52] = "R",
        [0x53] = "S",
        [0x54] = "T",
        [0x55] = "U",
        [0x56] = "V",
        [0x57] = "W",
        [0x58] = "X",
        [0x59] = "Y",
        [0x5A] = "Z",
        [0x5B] = "LWin",
        [0x5C] = "RWin",
        [0x5D] = "Apps",
        [0x5F] = "Sleep",
        [0x60] = "NumPad0",
        [0x61] = "NumPad1",
        [0x62] = "NumPad2",
        [0x63] = "NumPad3",
        [0x64] = "NumPad4",
        [0x65] = "NumPad5",
        [0x66] = "NumPad6",
        [0x67] = "NumPad7",
        [0x68] = "NumPad8",
        [0x69] = "NumPad9",
        [0x6A] = "Multiply",
        [0x6B] = "Add",
        [0x6C] = "Separator",
        [0x6D] = "Subtract",
        [0x6E] = "Decimal",
        [0x6F] = "Divide",
        [0x70] = "F1",
        [0x71] = "F2",
        [0x72] = "F3",
        [0x73] = "F4",
        [0x74] = "F5",
        [0x75] = "F6",
        [0x76] = "F7",
        [0x77] = "F8",
        [0x78] = "F9",
        [0x79] = "F10",
        [0x7A] = "F11",
        [0x7B] = "F12",
        [0x7C] = "F13",
        [0x7D] = "F14",
        [0x7E] = "F15",
        [0x7F] = "F16",
        [0x80] = "F17",
        [0x81] = "F18",
        [0x82] = "F19",
        [0x83] = "F20",
        [0x84] = "F21",
        [0x85] = "F22",
        [0x86] = "F23",
        [0x87] = "F24",
        [0x90] = "NumLock",
        [0x91] = "Scroll",
        [0x92] = "NumPadEnter",
        [0xA0] = "LShift",
        [0xA1] = "RShift",
        [0xA2] = "LControl",
        [0xA3] = "RControl",
        [0xA4] = "LMenu",
        [0xA5] = "RMenu",
        [0xBA] = ";",  -- OEM_1
        [0xBB] = "+",  -- OEM_Plus
        [0xBC] = ",",  -- OEM_Comma
        [0xBD] = "-",  -- OEM_Minus
        [0xBE] = ".",  -- OEM_Period
        [0xBF] = "/",  -- Slash
        [0xC0] = "`",  -- OEM_3
        [0xDB] = "[",  -- OEM_4
        [0xDC] = "\\", -- OEM_5 BackSlash
        [0xDD] = "]",  -- OEM_6
        [0xDE] = "'",  -- OEM_7
        -- [0xDF] = "OEM_8",
        -- [0xE2] = "OEM_102",
        -- [0xFE] = "DefinedEnter",
        -- [0xFF] = "DefinedCancel",
    }

    local key_name_to_code = {}
    for key, value in pairs(key_code_to_name) do
        key_name_to_code[value] = key
    end

    local reframework_key = {
        [0xBA] = true,  -- OEM_1 -> 0xBB
        [0xBB] = true,  -- OEM_Plus -> 0xDE
        [0xDE] = true,  -- OEM_7 -> 0xBA
        [0xC0] = true,  -- OEM_3 ` 不支持
        [0xDC] = true,  -- OEM_5 BackSlash 不支持
    }

    local pad_code_to_name = {
        [0x00000000] = "Not set",
        [0x00000001] = "LUp",
        [0x00000002] = "LDown",
        [0x00000004] = "LLeft",
        [0x00000008] = "LRight",
        [0x00000010] = "RUp",
        [0x00000020] = "RDown",
        [0x00000040] = "RLeft",
        [0x00000080] = "RRight",
        [0x00000100] = "LTrigTop",
        [0x00000200] = "LTrigBottom",
        [0x00000400] = "RTrigTop",
        [0x00000800] = "RTrigBottom",
        [0x00001000] = "LStickPush",
        [0x00002000] = "RStickPush",
        [0x00004000] = "CLeft",
        [0x00008000] = "CRight",
        [0x00010000] = "CCenter",
        [0x00020000] = "Decide",
        [0x00020020] = "RDown",
        [0x00040000] = "Cancel",
        [0x00040080] = "RRight",
        [0x00080000] = "PlatformHome",
        [0x00100000] = "EmuLup",
        [0x00200000] = "EmuLright",
        [0x00400000] = "EmuLdown",
        [0x00800000] = "EmuLleft",
        [0x01000000] = "EmuRup",
        [0x02000000] = "EmuRright",
        [0x04000000] = "EmuRdown",
        [0x08000000] = "EmuRleft",
        [0x10000000] = "LSide",
        [0x20000000] = "RSide",
        [0x40000000] = "RSL",
        [0x80000000] = "RSR",
    }


    local kb = sdk.call_native_func(keyboard_singleton, keyboard_typedef, "get_Device")
    local pad = sdk.call_native_func(gamePad_singleton, gamePad_typedef, "getMergedDevice", 0)
    local mouse = sdk.call_native_func(mouse_singleton, mouse_typedef, "get_Device")

    local function default_data()
        return { trigger_type = TRIGGER_TYPE_KEYBOARD }
    end

    local function is_valid(trigger_data)
        if not trigger_data or not trigger_data.trigger_type then
            return false
        end
        if trigger_data.trigger_type == TRIGGER_TYPE_KEYBOARD then
            return trigger_data.keycode ~= nil and trigger_data.keycode > 0
        elseif trigger_data.trigger_type == TRIGGER_TYPE_CONTROLLER then
            return trigger_data.pad ~= nil and trigger_data.pad > 0
        elseif trigger_data.trigger_type == TRIGGER_TYPE_MOUSE then
            return mouse_values[trigger_data.mouse] ~= nil
        end
        return false
    end

    -- 尝试设置按键
    local last_key_time = 0
    local function test_set_key(trigger_data)
        if not trigger_data then
            return false
        end
        if trigger_data.wait_set_key then
            local keycode = reframework:get_first_key_down()
            if keycode then
                if keycode == key_name_to_code.Escape then
                    keycode = nil
                end
                trigger_data.keycode = keycode
                trigger_data.wait_set_key = false
                last_key_time = os.clock()
            end
        elseif trigger_data.wait_set_pad then
            local pad_code = pad:get_ButtonRepeat()
            if pad_code ~= 0 then
                trigger_data.pad = pad_code
                trigger_data.wait_set_pad = false
                last_key_time = os.clock()
            end
        end
    end

    local _was_key_down = false
    local function reframework_is_key_down_once(code)
        local down = reframework:is_key_down(code)
        if not _was_key_down and down then
            _was_key_down = true
            return true
        end
        if not down then
            _was_key_down = false
        end
        return false
    end

    -- 检测是否按下
    local function is_trigger(trigger_data)
        if not is_valid(trigger_data) then
            return false
        end
        if os.clock() - last_key_time < 1 then
            return
        end
        if trigger_data.trigger_type == TRIGGER_TYPE_KEYBOARD then
            if trigger_data.keycode then
                local code = trigger_data.keycode
                if reframework_key[trigger_data.keycode] then
                    return reframework_is_key_down_once(code)
                elseif trigger_data.long_press then
                    return is_keyboard_long_press(kb, code)
                else
                    return kb:isRelease(code)
                end
            end
        elseif trigger_data.trigger_type == TRIGGER_TYPE_CONTROLLER then
            if trigger_data.pad then
                if trigger_data.long_press then
                    return is_pad_long_press(pad, trigger_data.pad)
                else
                    return (pad:get_ButtonUp() & trigger_data.pad) ~= 0
                end
            end
        elseif trigger_data.trigger_type == TRIGGER_TYPE_MOUSE then
            local mouse_code = mouse_values[trigger_data.mouse]
            if mouse_code then
                if trigger_data.long_press then
                    return is_mouse_long_press(mouse, mouse_code)
                else
                    return (mouse:get_ButtonUp() & mouse_code) ~= 0
                end
            end
        end
        return false
    end

    local message_dict_default = {
        trigger_with = "Trigger type",
        checked_set_key = "Checked to set a key, current: ",
        checked_set_controller_key = "Checked to set a controller key, current: ",
        choice_mouse_key = "Choice mouse key",
        need_long_press = "Need long press",
        unkown = "Unkown",
        unset = "Unset",
    }

    local message_dict_zh = {
        trigger_with = "触发方式",
        checked_set_key = "选中这里然后设置按键，当前: ",
        checked_set_controller_key = "选中这里设置手柄按键，当前: ",
        choice_mouse_key = "选择鼠标按键",
        need_long_press = "需要长按",
        unkown = "未知",
        unset = "未设置",
    }

    local function render_key_config(trigger_data, is_zh, label)
        local changed
        if label then
            if not imgui.tree_node(label) then return end
        end
        local message_dict = is_zh and message_dict_zh or message_dict_default
        changed, trigger_data.trigger_type = imgui.combo(
            message_dict.trigger_with,
            trigger_data.trigger_type, is_zh and trigger_type_choices_zh or trigger_type_choices)
        if trigger_data.trigger_type == TRIGGER_TYPE_KEYBOARD then
            changed, trigger_data.wait_set_key = imgui.checkbox(
                message_dict.checked_set_key ..
                    (trigger_data.keycode and (key_code_to_name[trigger_data.keycode] or tostring(trigger_data.keycode)) or message_dict.unset),
                trigger_data.wait_set_key)
        elseif trigger_data.trigger_type == TRIGGER_TYPE_CONTROLLER then
            changed, trigger_data.wait_set_pad = imgui.checkbox(
                message_dict.checked_set_controller_key
                    .. (pad_code_to_name[trigger_data.pad] or message_dict.unkown),
                trigger_data.wait_set_pad)
        elseif trigger_data.trigger_type == TRIGGER_TYPE_MOUSE then
            changed, trigger_data.mouse = imgui.combo(
                message_dict.choice_mouse_key,
                trigger_data.mouse,
                is_zh and mouse_choices_zh or mouse_choices)
        end
        changed, trigger_data.long_press = imgui.checkbox(
            message_dict.need_long_press,
            trigger_data.long_press)
        if label then
            imgui.tree_pop()
        end
    end

    -- local function render_key_bind_cfg()
    --     local changed, value
    --     changed, value = imgui.slider_float("Long press seconds", cfg.long_press_seconds, 0.2, 1, "%.1f")
    --     if changed then
    --         cfg.long_press_seconds = value
    --     end
    --     changed, value = imgui.slider_float("Long press suppress", cfg.long_press_suppress, 0, 1, "%.1f")
    --     tooltip("Long press repeat trigger suppress seconds")
    --     if changed then
    --         cfg.long_press_suppress = value
    --     end
    -- end

    return {
        TRIGGER_TYPE_KEYBOARD = TRIGGER_TYPE_KEYBOARD,
        TRIGGER_TYPE_CONTROLLER = TRIGGER_TYPE_CONTROLLER,
        TRIGGER_TYPE_MOUSE = TRIGGER_TYPE_MOUSE,
        kb,
        pad,
        mouse,
        default_data = default_data,
        test_set_key = test_set_key,
        is_valid = is_valid,
        is_trigger = is_trigger,
        render_key_config = render_key_config,
    }

end)()
local multi_language = (function ()

    -- imgui glyph ranges
    local GlyphRangesChineseFull = {
        0x0020, 0x00FF, -- Basic Latin + Latin Supplement
        0x2000, 0x206F, -- General Punctuation
        0x3000, 0x30FF, -- CJK Symbols and Punctuations, Hiragana, Katakana
        0x31F0, 0x31FF, -- Katakana Phonetic Extensions
        0xFF00, 0xFFEF, -- Half-width characters
        0xFFFD, 0xFFFD, -- Invalid
        0x4e00, 0x9FAF, -- CJK Ideograms
        0,
    }
    local GlyphRangesKorean = {
        0x0020, 0x00FF, -- Basic Latin + Latin Supplement
        0x3131, 0x3163, -- Korean alphabets
        0xAC00, 0xD7A3, -- Korean characters
        0xFFFD, 0xFFFD, -- Invalid
        0,
    }
    local GlyphRangesFrench = {
        0x0020, 0x017D,
        0,
    }
    local GlyphRangesRussian = {
        0x0020, 0x00FF,
        0x0400, 0x052f,
        0,
    }

    local optionManager = sdk.get_managed_singleton("chainsaw.OptionManager")
    local messageManager = sdk.get_managed_singleton("chainsaw.MessageManager")
    local MessageManager_typedef = sdk.find_type_definition("chainsaw.MessageManager")
    -- getMessage(System.Guid message_id, via.Language language_id, bool emptyable)
    local getMessage_method = MessageManager_typedef:get_method("getMessage(System.Guid, via.Language, System.Boolean)")

    local export = {
        Language_English = 0x1,
        Language_SimplelifiedChinese = 0xD,
        Language_TransitionalChinese = 0xC,
        Language_Japanese = 0x0,
        Language_French = 0x2,
        Language_Russian = 0x6,
        Language_Korean = 0xB,
        Language_Arabic = 0x15,
        language = nil,
        font = nil,
    }

    local last_clock = 0

    function export.handle_language(cfg, on_change, force_reload)
        local now = os.clock()
        if now - last_clock < 0.2 then
            return
        end
        last_clock = now
        local new_language
        if cfg.use_english then
            new_language = export.Language_English
            export.font = nil
        else
            new_language = optionManager:getTextLanguage()
        end
        local language = export.language
        if new_language == language and not force_reload then
            return
        end
        language = new_language
        export.language = language
        local is_zh = false
        if language == export.Language_SimplelifiedChinese or
            language == export.Language_TransitionalChinese then
            is_zh = true
        end
        local font
        if not cfg.builtin_font then
            if language == export.Language_SimplelifiedChinese or
                language == export.Language_TransitionalChinese or
                language == export.Language_Japanese then
                font = imgui.load_font("STXIHEI.TTF", cfg.font_size, GlyphRangesChineseFull)
            elseif language == export.Language_French then
                font = imgui.load_font("STXIHEI.TTF", cfg.font_size, GlyphRangesFrench)
            elseif language == export.Language_Russian then
                font = imgui.load_font("STXIHEI.TTF", cfg.font_size, GlyphRangesRussian)
            elseif language == export.Language_Korean then
                font = imgui.load_font("malgunsl.ttf", cfg.font_size, GlyphRangesKorean)
            end
        end
        export.font = font
        if on_change then
            on_change(is_zh)
        end
    end

    function export.getMessage(msdId)
        return getMessage_method:call(messageManager, msdId, export.language, true)
    end

    return export

end)()
local Mod = (function ()
    local Mod = {}

    local common_cfg = {
        use_english = false,
        builtin_font = false,
        font_size = 30,
    }

    local common_message_dict_default = {
        reset_config = "Reset to default",
        save_config = "Save config",
        use_english = "Use English",
        builtin_font = "Use builtin font",
        reload_message = "Reload info (when language changed)",
        font_size = "Font size (20~60)",
        hot_key = " Hotkey",
    }

    local common_message_dict_zh = {
        font_size = "字体大小 (20~60)",
        reset_config = "重置配置",
        save_config = "保存配置",
        builtin_font = "使用内建字体",
        reload_message = "刷新数据(切换语言后)",
        hot_key = "快捷键",
    }

    local meta_return_key = { __index = function(mytable, key) return key end }

    function Mod:new(title, cfg_path, default_cfg, message_dict_default, message_dict_zh)
        local obj = {
            title = title,
            cfg_path = cfg_path,
            default_cfg = default_cfg,
            message_dict_default = message_dict_default,
            message_dict_zh = message_dict_zh,
            message_dict = message_dict_default,
            is_zh = false,  -- 当前语言是中文
            enable_font_size = true,
            render_window = false,
            zh_title = false,  -- 标题使用中文字体
            handle_language_callback = nil,
        }
        setmetatable(obj, self)
        self.__index = self

        Mod.table_assign_default(message_dict_default, common_message_dict_default)
        if message_dict_zh then
            setmetatable(message_dict_zh, message_dict_default)
            Mod.table_assign_default(message_dict_zh, common_message_dict_zh)
        end
        message_dict_default.__index = message_dict_default
        setmetatable(message_dict_default, meta_return_key)

        local cfg = json.load_file(cfg_path)
        if type(cfg) ~= 'table' then
            cfg = {}
        end
        Mod.table_assign_default(cfg, common_cfg)
        Mod.table_assign_default(cfg, default_cfg)
        obj.cfg = cfg

        re.on_config_save(function()
            obj:dump_cfg()
        end)

        return obj
    end

    function Mod.table_assign_default(dst, data)
        for key, value in pairs(data) do
            if dst[key] == nil then
                if type(value) == 'table' then
                    dst[key] = Mod.table_assign(nil, value)
                else
                    dst[key] = value
                end
            end
        end
        return dst
    end

    function Mod.table_assign(dst, data)
        dst = dst or {}
        if data == nil then
            return dst
        end
        for key, value in pairs(data) do
            if type(value) == 'table' then
                dst[key] = Mod.table_assign(nil, value)
            else
                dst[key] = value
            end
        end
        return dst
    end

    function Mod.empty_or(text, default)
        if text == nil or text == "" then
            return default
        end
        return text
    end

    function Mod:dump_cfg()
        json.dump_file(self.cfg_path, self.cfg)
    end

    function Mod:init_gui_ids()
        self.gui_ids = setmetatable({__cur = 0}, {
            __index = function (data, name)
                local index = data.__cur + 1
                data[name] = index
                data.__cur = index
                return index
            end
        })
    end

    function Mod:init_multi_language(multi_language)
        self.multi_language = multi_language
    end

    function Mod:init_key_bind(key_bind)
        self.key_bind = key_bind
    end

    function Mod:on_language_changed(is_zh)
        self.is_zh = is_zh
        self.message_dict = is_zh and self.message_dict_zh or self.message_dict_default
    end

    function Mod:handle_language(force_reload)
        if self.multi_language then
            self.multi_language.handle_language(self.cfg, function (is_zh)
                self:on_language_changed(is_zh)
                if self.handle_language_callback then
                    self.handle_language_callback(is_zh)
                end
            end, force_reload)
        end
    end

    function Mod:text(name)
        return self.message_dict[name]
    end

    function Mod.tooltip_text(text)
        if imgui.is_item_hovered() then
            imgui.set_tooltip(text)
        end
    end

    function Mod:tooltip(name)
        if imgui.is_item_hovered() then
            imgui.set_tooltip(self.message_dict[name])
        end
    end

    function Mod:render_checkbox(name)
        local changed, value = imgui.checkbox(self.message_dict[name], self.cfg[name])
        if changed then
            self.cfg[name] = value
        end
        return changed, value
    end

    function Mod:render_slider_float(name, min, max, flag)
        local changed, value = imgui.slider_float(self.message_dict[name], self.cfg[name], min, max, flag)
        if changed then
            self.cfg[name] = value
        end
        return changed, value
    end

    function Mod:render_color_edit(name)
        local changed, value = imgui.color_edit(self.message_dict[name], self.cfg[name])
        if changed then
            self.cfg[name] = value
        end
    end

    function Mod:render_set_key(name, suffix)
        if not self.key_bind then
            return
        end
        local trigger_key = name .. (suffix or "HotKey")
        if not self.cfg[trigger_key] then
            self.cfg[trigger_key] = self.key_bind.default_data()
        end
        if self.message_dict[trigger_key] == trigger_key then
            self.message_dict[trigger_key] = self:text(name) .. self:text("hot_key")
        end
        self.key_bind.render_key_config(self.cfg[trigger_key], self.is_zh, self.message_dict[trigger_key])
    end

    -- set and check key trigger
    function Mod:handle_key(trigger_data)
        if not self.key_bind or not trigger_data then
            return
        end
        self.key_bind.test_set_key(trigger_data)
        return self.key_bind.is_trigger(trigger_data)
    end

    -- hotkey toggle checkbox
    function Mod:handle_key_toggle(name)
        if not self.key_bind then
            return
        end
        local trigger_key = name .. "HotKey"
        self.key_bind.test_set_key(self.cfg[trigger_key])
        if self.key_bind.is_trigger(self.cfg[trigger_key]) then
            self.cfg[name] = not self.cfg[name]
        end
    end

    function Mod:push_font()
        if self.multi_language then
            if not self.multi_language.language then
                self:handle_language()
            end
            local cur_font = self.multi_language.font
            if cur_font then
                imgui.push_font(cur_font)
                return true
            end
        end
        return false
    end

    function Mod:begin_render_config(shown_flag)
        imgui.set_next_item_open(true, 1 << 1)
        local title = self.title or self.message_dict.window_title
        local has_font = self.zh_title and self:push_font()

        if self.render_window then
            if not imgui.begin_window(title, shown_flag ~= false, 1 << 10) then
                if has_font then
                    imgui.pop_font()
                end
                return false
            end
        else
            if not imgui.tree_node(title) then
                if has_font then
                    imgui.pop_font()
                end
                return false
            end
        end
        if not self.zh_title then
            self:push_font()
        end
        local changed, value
        local use_english_changed, builtin_font_changed
        use_english_changed, value = self:render_checkbox("use_english")
        imgui.same_line()
        builtin_font_changed, value = self:render_checkbox("builtin_font")
        if use_english_changed or builtin_font_changed then
            self:handle_language(true)
        end
        if self.is_zh then
            imgui.text("RE4修改和mod交流群: 663068787")
        end
        if imgui.button(self.message_dict.reload_message) then
            self:handle_language(true)
        end
        if self.enable_font_size and not self.cfg.builtin_font then
            imgui.same_line()
            imgui.set_next_item_width(200)
            changed, value = imgui.input_text(self.message_dict.font_size, self.cfg.font_size)
            if changed then
                value = tonumber(value)
                if value and value >= 20 and value <= 60 then
                    self.cfg.font_size = value
                    self:handle_language(true)
                end
            end
        end
        if imgui.button(self.message_dict.reset_config) then
            Mod.table_assign(self.cfg, self.default_cfg)
        end
        imgui.same_line()
        if imgui.button(self.message_dict.save_config) then
            self:dump_cfg()
        end

        return true
    end

    function Mod:end_render_config()
        local cur_font = self.multi_language and self.multi_language.font
        if cur_font then
            imgui.pop_font()
        end
        if self.render_window then
            imgui.end_window()
        else
            imgui.tree_pop()
        end
    end

    return Mod

end)()

local cfg_path = "item_indicator/config.json"
local default_cfg = {
    font_size = 30,
    use_english = false,
    show_item = true,
    show_normal = true,
    show_key = true,
    show_treasure = true,
    show_sleep = true,
    show_box = true,
    show_blue_coin = true,
    limit_distance = 50.0,
    auto_pick = false,
    auto_pick_distance = 10.0,
    color_normal_item = 0xFFFFFFFF,
    color_key_item = 0xFF0000FF,
    color_treasure_item = 0xFF00D7FF,
    color_shoot = 0xFFE99237,
    scan_interval = 1.0,
    show_item_trigger = {
        trigger_type = key_bind.TRIGGER_TYPE_KEYBOARD,
        keycode = 35,
        long_press = false,
    },
    auto_pick_trigger = {},
}

local message_dict_default = {
    show_item = "Show indicator",
    show_normal = "Show normal item",
    show_key = "Show key item",
    show_treasure = "Show treasure item",
    show_sleep = "Show sleep item",
    box = "Box",
    show_box = "Show box",
    blue_coin = "Blue medallions",
    show_blue_coin = "Show blue medallions and Clockwork Castellan",
    sr_ramon = "Clockwork Castellan",
    limit_distance = "Visiable distance",
    scan_interval = "Scan interval seconds",
    auto_pick = "Enable auto pick",
    auto_pick_distance = "Auto pick distance",
    config_color = "Config text color",
    color_normal_item = "Normal item color",
    color_key_item = "Key item color",
    color_treasure_item = "Treasure item color",
    color_shoot = "Blue medallions and Clockwork Castellan color",
}
local message_dict_zh = {
    show_item = "显示指示器",
    show_normal = "显示常规物品",
    show_key = "显示关键物品",
    show_treasure = "显示珍宝",
    show_sleep = "显示沉睡物品",
    box = "箱子",
    show_box = "显示箱子",
    blue_coin = "蓝牌",
    sr_ramon = "城主人偶",
    show_blue_coin = "显示蓝牌和城主人偶",
    limit_distance = "可视距离",
    scan_interval = "扫描间隔秒数",
    auto_pick = "自动拾取",
    auto_pick_distance = "自动拾取距离",
    config_color = "设置文字颜色",
    color_normal_item = "普通物品颜色",
    color_key_item = "关键物品颜色",
    color_treasure_item = "珍宝颜色",
    color_shoot = "蓝牌和城主人偶颜色",
}

local mod = Mod:new("Item indicator", cfg_path, default_cfg, message_dict_default, message_dict_zh)
mod:init_multi_language(multi_language)
mod:init_key_bind(key_bind)
local cfg = mod.cfg

local item_msg = {}
mod.handle_language_callback = function(is_zh)
    item_msg = {}
end

local last_time = 0
-- local scene
local player
local player_position
local items = {}
local now

local function is_ready()
    return player_position ~= nil
end

-- local function get_scene()
--     return sdk.call_native_func(sceneManager, SceneManager_typedef, "get_CurrentScene()")
-- end

re.on_frame(function ()
    if not (cfg.show_item or cfg.auto_pick) then
        return
    end
    now = os.clock()
    if now - last_time > cfg.scan_interval then
        if not multi_language.language then
            mod:handle_language()
        end
        last_time = now
        -- scene = get_scene()
        player = characterManager:getPlayerContextRef()
        if player then
            player_position = player:get_Position()
        end
    end
    if not cfg.show_item or not is_ready() then return end
    local cur_font = multi_language.font
    if cur_font then
        imgui.push_font(cur_font)
    end
    for key, item in pairs(items) do
        if item and item.display_text then
            local color = cfg.color_normal_item
            if item.is_key then
                color = cfg.color_key_item
            elseif item.is_treasure then
                color = cfg.color_treasure_item
            elseif item.is_shoot then
                color = cfg.color_shoot
            end
            draw.world_text(item.display_text, item.pos, color)
            if now - item.update_time > cfg.scan_interval then
                items[key] = nil
            end
        end
    end
    if cur_font then
        imgui.pop_font()
    end
end)

local function pickup_item(drop_item)
    if not player then return end
    local bodyUpdater = player._BodyUpdater
    if math.abs(bodyUpdater:get_MoveDir()) < 0.000001 then
        return
    end
    local mfsm2 = bodyUpdater:get_MotionFsm()
    if mfsm2 then
        local current_name = mfsm2:call("getCurrentNodeName", 0)
        if not (current_name:find("^Run%.") or current_name:find("^Walk%.") or current_name:find("^Idle%.")) then
            return
        end
    end
    if isInventoryItem:call(nil, drop_item:getItemID()) then
        local player_HeadUpdater = player:get_HeadUpdater()
        local controller = player_HeadUpdater:get_InventoryController()
        if not controller:call("enableAddItem(chainsaw.Item, chainsaw.AddItemType)", drop_item._Item, 0) then
            return
        end
    end
    drop_item:onAcceptPickup(player:get_SpawnerID(), drop_item._ID)
    -- log.debug("item_id: " .. tostring(drop_item._ID._Code))
end

local function setup_item(item)
    if item.name then
        item.display_text = string.format("%s: %.0f", item.name, item.distance)
    end
    item.update_time = now
    return item
end

local function update_item_time(addr)
    if items[addr] then
        items[addr].update_time = now
    else
        items[addr] = { update_time = now }
    end
end

local function need_update(userdata)
    local addr = sdk.to_int64(userdata)
    local item = items[addr]
    local result = not item or now - item.update_time > cfg.scan_interval
    if result then
        update_item_time(addr)
    end
    return result
end

local function add_item(item, auto_pick)
    local addr = item:get_address()
    local pos = item:getIconPos()
    local distance = (player_position - pos):length()
    if distance <= cfg.limit_distance then
        local item_id = item:getItemID()
        local is_key = isKeyItem:call(nil, item_id)
        if cfg.show_item then
            local item_name
            if not item_msg[item_id] then
                local msdId = itemMessageManager:getItemNameMsgId(item_id)
                item_name = multi_language.getMessage(msdId)
                item_msg[item_id] = item_name
            else
                item_name = item_msg[item_id]
            end
            if pos.y > player_position.y + 4 then
                item_name = item_name .. "^"
            elseif pos.y < player_position.y - 4 then
                item_name = item_name .. "_"
            end
            local is_treasure = isTreasureItem:call(nil, item_id)
            if (cfg.show_key and is_key) or
                (cfg.show_treasure and is_treasure) or
                (cfg.show_normal and not (is_key or is_treasure)) then
                items[addr] = setup_item({
                    pos = pos,
                    name = item_name,
                    distance = distance,
                    is_key = is_key,
                    is_treasure = is_treasure,
                })
            end
        end
        if auto_pick and cfg.auto_pick and distance <= cfg.auto_pick_distance and not is_key then
            pickup_item(item)
        end
    end
end

sdk.hook(
    DropItem_typedef:get_method("cantDoubleBilling"),
    function (args)
        if not (cfg.show_item or cfg.auto_pick) or not is_ready() then
            return
        end
        if need_update(args[2]) then
            local item = sdk.to_managed_object(args[2])
            if item._Item then
                add_item(item, true)
            end
        end
    end
)

local PTAS_ID = 124000000
local DIRTY_PEARL_PANDENT_ID = 120803200

sdk.hook(
    DropItem_typedef:get_method("updateSleep"),
    function (args)
        if not (cfg.show_item and cfg.show_sleep) or not is_ready() then
            return
        end
        if need_update(args[2]) then
            local item = sdk.to_managed_object(args[2])
            local item_id = item:getItemID()
            if item_id ~= PTAS_ID and item_id ~= DIRTY_PEARL_PANDENT_ID then
                add_item(item, false)
            end
        end
    end
)

local GameObject_typedef = sdk.find_type_definition("via.GameObject")
local get_Transform = GameObject_typedef:get_method("get_Transform")

sdk.hook(
    GmSmoothWoodBox_typedef:get_method("updateWait"),
    function (args)
        if not (cfg.show_item and cfg.show_box) or not is_ready() then
            return
        end
        if not need_update(args[2]) then
            return
        end
        local box = sdk.to_managed_object(args[2])
        if not box then
            return
        end
        local object = box._ObjAfter
        if object then
            local transform = get_Transform:call(object)
            local pos = transform:get_Position()
            local distance = (player_position - pos):length()
            if distance <= cfg.limit_distance then
                local name = mod.message_dict.box
                if not name then
                    return
                end
                if pos.y > player_position.y + 4 then
                    name = name .. "^"
                elseif pos.y < player_position.y - 4 then
                    name = name .. "_"
                end
                items[box:get_address()] = setup_item({
                    pos = pos,
                    distance = distance,
                    name = name,
                })
            end
        end
    end
)

sdk.hook(
    GmShootObjBase_typedef:get_method("updateWait"),
    function (args)
        if not (cfg.show_item and cfg.show_blue_coin) or not is_ready() then
            return
        end
        if not need_update(args[2]) then
            return
        end
        local gm = sdk.to_managed_object(args[2])
        if not gm then
            return
        end
        local object = gm._BreakObj
        if object then
            local transform = get_Transform:call(object)
            local pos = transform:get_Position()
            local distance = (player_position - pos):length()
            if distance <= cfg.limit_distance then
                local type = gm:get_type_definition()
                local name
                if type == GmBlueCoin_typedef then
                    name = mod.message_dict.blue_coin
                elseif type == GmSrRamon_typedef then
                    name = mod.message_dict.sr_ramon
                end
                if not name then
                    return
                end
                if pos.y > player_position.y + 4 then
                    name = name .. "^"
                elseif pos.y < player_position.y - 4 then
                    name = name .. "_"
                end
                items[gm:get_address()] = setup_item({
                    name = name,
                    pos = pos,
                    distance = distance,
                    is_shoot = true
                })
            end
        end
    end
)

re.on_application_entry("UpdateHID", function()
    if mod:handle_key(cfg.show_item_trigger) then
        cfg.show_item = not cfg.show_item
    end
    if mod:handle_key(cfg.auto_pick_trigger) then
        cfg.auto_pick = not cfg.auto_pick
    end
end)

local function render_set_key(name)
    mod:render_set_key(name, "_trigger")
end

re.on_draw_ui(function()
    local changed, value
    imgui.set_next_item_open(true, 1 << 1)
    if mod:begin_render_config() then
        mod:render_checkbox("show_item")
        mod:render_checkbox("show_normal")
        mod:render_checkbox("show_key")
        mod:render_checkbox("show_treasure")
        mod:render_checkbox("show_sleep")
        mod:render_checkbox("show_box")
        mod:render_checkbox("show_blue_coin")
        mod:render_slider_float("limit_distance", 5, 100, "%.1f")
        mod:render_slider_float("scan_interval", 1, 5, "%.1f")
        mod:render_checkbox("auto_pick")
        if cfg.auto_pick then
            mod:render_slider_float("auto_pick_distance", 1, 15, "%.1f")
        end
        if imgui.tree_node(mod.message_dict.config_color) then
            mod:render_color_edit("color_normal_item")
            mod:render_color_edit("color_key_item")
            mod:render_color_edit("color_treasure_item")
            mod:render_color_edit("color_shoot")
            imgui.tree_pop()
        end

        render_set_key("show_item")
        render_set_key("auto_pick")
        mod:end_render_config()
    end
end)
